#!/usr/bin/env python3

import re
import sys
import os
import argparse
sys.path.append(os.path.join(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
    "lib"
))
from tgs.exporters import exporters
from tgs.importers import importers
from tgs.utils.stripper import float_strip, heavy_strip
from tgs import __version__


desc = """Converts between multiple formats

Supported formats:

- Input:
%s

- Output:
%s
""" % (
    "\n".join("%s- %s" % (" "*2, e.name) for e in importers),
    "\n".join("%s- %s" % (" "*2, e.name) for e in exporters),
)

parser = argparse.ArgumentParser(
    description=desc,
    formatter_class=argparse.RawDescriptionHelpFormatter,
    conflict_handler='resolve'
)
parser.add_argument("--version", "-v", action="version", version="%(prog)s - tgs " + __version__)

group = importers.set_options(parser)
group.add_argument(
    "infile",
    default="-",
    nargs="?",
    help="Input file"
)
group.add_argument(
    "--input-format", "-if",
    default=None,
    choices=[importer.slug for importer in importers],
    help="Explicit output format (if missing implied by the input filename)",
)


group = exporters.set_options(parser)

group.add_argument(
    "outfile",
    default="-",
    nargs="?",
    help="Output file"
)
group.add_argument(
    "--output-format", "-of",
    default=None,
    choices=[exporter.slug for exporter in exporters],
    help="Explicit output format (if missing implied by the output filename)",
)
group.add_argument(
    "--sanitize",
    default=False,
    action="store_true",
    help="Ensure the animation is 512x512 and 30 or 60 fps",
)
group.add_argument(
    "--optimize",
    "-O",
    default=1,
    type=int,
    choices=[0, 1, 2],
    help="Optimize the animation parameter: 0 no optimization, 1 truncate floats, 2 truncate floats and names",
)

if __name__ == "__main__":
    ns = parser.parse_args()

    infile = ns.infile
    importer = None
    if infile == "-":
        infile = sys.stdin
    else:
        suf = os.path.splitext(infile)[1][1:]
        for p in importers:
            if suf in p.extensions:
                importer = p
                break
    if ns.input_format:
        importer = None
        for p in importers:
            if p.slug == ns.input_format:
                importer = p
                break
    if not importer:
        sys.stderr.write("Unknown importer\n")
        sys.exit(1)

    outfile = ns.outfile
    exporter = None
    if outfile == "-":
        outfile = sys.stdout
    else:
        exporter = exporters.get_from_filename(outfile)
    if ns.output_format:
        exporter = exporters.get(ns.output_format)

    if not exporter:
        sys.stderr.write("Unknown exporter\n")
        sys.exit(1)

    i_options = {}
    for opt in importer.extra_options:
        i_options[opt.name] = getattr(ns, opt.nsvar(importer.slug))

    o_options = exporter.argparse_options(ns)

    an = importer.process(infile, **i_options)
    if ns.optimize == 1:
        float_strip(an)
    elif ns.optimize >= 2:
        heavy_strip(an)
    if ns.sanitize:
        an.tgs_sanitize()
    exporter.process(an, outfile, **o_options)



from .base import TgsObject, TgsProp, TgsEnum
from .properties import MultiDimensional, Value, NVector, ShapeProperty


## \ingroup Lottie
class Transform(TgsObject):
    """!
    Layer transform
    """
    _props = [
        TgsProp("anchor_point", "a", MultiDimensional, False),
        TgsProp("position", "p", MultiDimensional, False),
        TgsProp("scale", "s", MultiDimensional, False),
        TgsProp("rotation", "r", Value, False),
        TgsProp("opacity", "o", Value, False),
        #TgsProp("position_x", "px", Value, False),
        #TgsProp("position_y", "py", Value, False),
        #TgsProp("position_z", "pz", Value, False),
        TgsProp("skew", "sk", Value, False),
        TgsProp("skew_axis", "sa", Value, False),
    ]

    def __init__(self):
        ## Transform Anchor Point
        self.anchor_point = MultiDimensional(NVector(0, 0))
        ## Transform Position
        self.position = MultiDimensional(NVector(0, 0))
        ## Transform Scale
        self.scale = MultiDimensional(NVector(100, 100))
        ## Transform Rotation
        self.rotation = Value(0)
        ## Transform Opacity
        self.opacity = Value(100)

        """
        # Transform Position X
        #self.position_x = Value()
        ## Transform Position Y
        #self.position_y = Value()
        ## Transform Position Z
        #self.position_z = Value()
        """

        ## Transform Skew
        self.skew = Value(0)
        ## Transform Skew Axis.
        ## An angle, if 0 skews on the X axis, if 90 skews on the Y axis
        self.skew_axis = Value(0)


## \ingroup Lottie
class MaskMode(TgsEnum):
    """!
    How masks interact with each other
    \see https://helpx.adobe.com/after-effects/using/alpha-channels-masks-mattes.html
    """
    No = "n"
    Add = "a"
    Subtract = "s"
    Intersect = "i"
    ## @note Not in lottie web
    Lightent = "l"
    ## @note Not in lottie web
    Darken = "d"
    ## @note Not in lottie web
    Difference = "f"


## \ingroup Lottie
## \todo Implement SVG/SIF I/O
class Mask(TgsObject):
    _props = [
        TgsProp("inverted", "inv", bool, False),
        TgsProp("name", "nm", str, False),
        TgsProp("shape", "pt", ShapeProperty, False),
        TgsProp("opacity", "o", Value, False),
        TgsProp("mode", "mode", MaskMode, False),
        TgsProp("dilate", "x", Value, False),
    ]

    def __init__(self, bezier=None):
        ## Inverted Mask flag
        self.inverted = False
        ## Mask name. Used for expressions and effects.
        self.name = None
        ## Mask vertices
        self.shape = ShapeProperty(bezier)
        ## Mask opacity.
        self.opacity = Value(100)
        ## Mask mode. Not all mask types are supported.
        self.mode = MaskMode.Intersect
        self.dilate = Value(0)



import enum
import inspect
import importlib
from ..nvector import NVector


class Tgs:
    """!
    Base class for TGS/Lottie JSON objects bindings
    """
    def to_dict(self):
        """!
        Serializes into a JSON object fit for the Lottie format
        """
        raise NotImplementedError

    @classmethod
    def load(cls, lottiedict):
        """!
        Loads from a JSON object
        \returns An instance of the class
        """
        raise NotImplementedError

    def clone(self):
        """!
        Returns a copy of the object
        """
        raise NotImplementedError


class EnumMeta(enum.EnumMeta):
    """!
    Hack to counter-hack the hack in enum meta
    """
    def __new__(cls, name, bases, classdict):
        classdict["__reduce_ex__"] = lambda *a, **kw: None  # pragma: no cover
        return super().__new__(cls, name, bases, classdict)


class TgsEnum(Tgs, enum.Enum, metaclass=EnumMeta):
    """!
    Base class for enum-like types in the Lottie JSON structure
    """
    def to_dict(self):
        return self.value

    @classmethod
    def load(cls, lottieint):
        return cls(lottieint)

    def clone(self):
        return self


class PseudoList:
    """!
    List tag for some weird values in the Lottie JSON
    """
    pass


class TgsConverter:
    """!
    Factory for property types that require special conversions
    """
    def __init__(self, py, lottie, name=None):
        self.py = py
        self.lottie = lottie
        self.name = name or "%s but displayed as %s" % (self.py.__name__, self.lottie.__name__)

    def py_to_lottie(self, val):
        return self.lottie(val)

    def lottie_to_py(self, val):
        return self.py(val)

    @property
    def __name__(self):
        return self.name


## For values in Lottie that are bools but ints in the JSON
PseudoBool = TgsConverter(bool, int, "0-1 int")


class TgsProp:
    """!
    Lottie <-> Python property mapper
    """
    def __init__(self, name, lottie, type=float, list=False, cond=None):
        ## Name of the Python property
        self.name = name
        ## Name of the Lottie JSON property
        self.lottie = lottie
        ## Type of the property
        ## \see TgsConverter, PseudoBool
        self.type = type
        ## Whether the property is a list of self.type
        ## \see PseudoList
        self.list = list
        ## Condition on when the property is loaded from the Lottie JSON
        self.cond = cond

    def get(self, obj):
        """!
        Returns the value of the property from a Python object
        """
        return getattr(obj, self.name)

    def set(self, obj, value):
        """!
        Sets the value of the property from a Python object
        """
        if isinstance(getattr(obj.__class__, self.name, None), property):
            return
        return setattr(obj, self.name, value)

    def load_from_parent(self, lottiedict):
        """!
        Returns the value for this property from a JSON dict representing the parent object
        \returns The loaded value or \c None if the property is not in \p lottiedict
        """
        if self.lottie in lottiedict:
            return self.load(lottiedict[self.lottie])
        return None

    def load_into(self, lottiedict, obj):
        """!
        Loads from a Lottie dict into an object
        """
        if self.cond and not self.cond(lottiedict):
            return
        self.set(obj, self.load_from_parent(lottiedict))

    def load(self, lottieval):
        """!
        Loads the property from a JSON value
        \returns the Python equivalent of the JSON value
        """
        if self.list is PseudoList and isinstance(lottieval, list):
            return self._load_scalar(lottieval[0])
            #return [
                #self._load_scalar(it)
                #for it in lottieval
            #]
        elif self.list is True:
            return [
                self._load_scalar(it)
                for it in lottieval
            ]
        return self._load_scalar(lottieval)

    def _load_scalar(self, lottieval):
        try:
            if inspect.isclass(self.type) and issubclass(self.type, Tgs):
                return self.type.load(lottieval)
            elif isinstance(self.type, type) and isinstance(lottieval, self.type):
                return lottieval
            elif isinstance(self.type, TgsConverter):
                return self.type.lottie_to_py(lottieval)
            elif self.type is NVector:
                return NVector(*lottieval)
            return self.type(lottieval)
        except Exception as e:
            raise TypeError("Could not load `%s` (%s) as %s:\n%s: %s" % (
                self.lottie,
                self.name,
                self.type.__name__,
                e.__class__.__name__,
                e
            )) from e

    def to_dict(self, obj):
        """!
        Converts the value of the property as from \p obj into a JSON value
        \param obj TgsObject with this property
        """
        val = self._basic_to_dict(self.get(obj))
        if self.list is PseudoList:
            if not isinstance(obj, list):
                return [val]
        elif isinstance(self.type, TgsConverter):
            val = self._basic_to_dict(self.type.py_to_lottie(val))
        return val

    def _basic_to_dict(self, v):
        try:
            if isinstance(v, Tgs):
                return v.to_dict()
            elif isinstance(v, NVector):
                return list(map(self._basic_to_dict, v.components))
            elif isinstance(v, list):
                return list(map(self._basic_to_dict, v))
            elif isinstance(v, (int, str, bool)):
                return v
            elif isinstance(v, float):
                if v % 1 == 0:
                    return int(v)
                return v #round(v, 3)
            else:
                raise Exception("Unknown value %r" % v)
        except Exception as e:
            raise TypeError("Could not save `%s` (%s) as %s:\n%s: %s" % (
                self.lottie,
                self.name,
                self.type.__name__,
                e.__class__.__name__,
                e
            )) from e

    def __repr__(self):
        return "<TgsProp %s:%s>" % (self.name, self.lottie)

    def clone_value(self, value):
        if isinstance(value, list):
            return [self.clone_value(v) for v in value]
        if isinstance(value, (Tgs, NVector)):
            return value.clone()
        if isinstance(value, (int, float, bool, str)) or value is None:
            return value
        raise Exception("Could not convert %r" % value)


class TgsObjectMeta(type):
    def __new__(cls, name, bases, attr):
        props = []
        for base in bases:
            if type(base) == cls:
                props += base._props
        attr["_props"] = props + attr.get("_props", [])
        return super().__new__(cls, name, bases, attr)


class TgsObject(Tgs, metaclass=TgsObjectMeta):
    """!
    @brief Base class for mapping Python classes into Lottie JSON objects
    """
    def to_dict(self):
        return {
            prop.lottie: prop.to_dict(self)
            for prop in self._props
            if prop.get(self) is not None
        }

    @classmethod
    def load(cls, lottiedict):
        if "__pyclass" in lottiedict:
            return CustomObject.load(lottiedict)
        cls = cls._load_get_class(lottiedict)
        obj = cls()
        for prop in cls._props:
            prop.load_into(lottiedict, obj)
        return obj

    @classmethod
    def _load_get_class(cls, lottiedict):
        return cls

    def find(self, search, propname="name"):
        """!
        \param search   The value of the property to search
        \param propname The name of the property used to search
        @brief Recursively searches for child objects with a matching property
        """
        if getattr(self, propname, None) == search:
            return self
        for prop in self._props:
            v = prop.get(self)
            if isinstance(v, TgsObject):
                found = v.find(search, propname)
                if found:
                    return found
            elif isinstance(v, list) and v and isinstance(v[0], TgsObject):
                for obj in v:
                    found = obj.find(search, propname)
                    if found:
                        return found
        return None

    def clone(self):
        obj = self.__class__()
        for prop in self._props:
            v = prop.get(self)
            prop.set(obj, prop.clone_value(v))
        return obj


class Index:
    """!
    @brief Simple iterator to generate increasing integers
    """
    def __init__(self):
        self._i = -1

    def __next__(self):
        self._i += 1
        return self._i


class CustomObject(TgsObject):
    """!
    Allows extending the Lottie shapes with custom Python classes
    """
    wrapped_tgs = TgsObject

    def __init__(self):
        self.wrapped = self.wrapped_tgs()

    @classmethod
    def load(cls, lottiedict):
        ld = lottiedict.copy()
        classname = ld.pop("__pyclass")
        modn, clsn = classname.rsplit(".", 1)
        subcls = getattr(importlib.import_module(modn), clsn)
        obj = subcls()
        for prop in subcls._props:
            prop.load_into(lottiedict, obj)
        obj.wrapped = subcls.wrapped_tgs.load(ld)
        return obj

    def clone(self):
        obj = self.__class__(**self.to_pyctor())
        obj.wrapped = self.wrapped.clone()
        return obj

    def to_dict(self):
        dict = self.wrapped.to_dict()
        dict["__pyclass"] = "{0.__module__}.{0.__name__}".format(self.__class__)
        dict.update(TgsObject.to_dict(self))
        return dict

    def _build_wrapped(self):
        return self.wrapped_tgs()

    def refresh(self):
        self.wrapped = self._build_wrapped()


class ObjectVisitor:
    DONT_RECURSE = object()

    def __call__(self, tgs_object):
        self._process(tgs_object)

    def _process(self, tgs_object):
        self.visit(tgs_object)
        for p in tgs_object._props:
            pval = p.get(tgs_object)
            if self.visit_property(tgs_object, p, pval) is not self.DONT_RECURSE:
                if isinstance(pval, TgsObject):
                    self._process(pval)
                elif isinstance(pval, list) and pval and isinstance(pval[0], TgsObject):
                    for c in pval:
                        self._process(c)

    def visit(self, object):
        pass

    def visit_property(self, object, property, value):
        pass



from .base import importer
from ..parsers.baseporter import ExtraOption
from ..parsers.pixel import pixel_to_animation_paths, pixel_to_animation
from ..parsers.svg.importer import parse_color

try:
    from ..parsers.raster import raster_to_animation
    raster = True
except ImportError:
    raster = False


@importer("Raster image", ["bmp", "png", "gif", "webp", "tiff"], [
    ExtraOption("n_colors", type=int, default=1, help="Number of colors to quantize"),
    ExtraOption("palette", type=parse_color, default=[], nargs="+", help="Custom palette"),
    ExtraOption(
        "mode",
        default="pixel",
        choices=["pixel", "polygon"] + (["bezier"] if raster else []),
        help="Vectorization mode"
    ),
    ExtraOption("frame_delay", type=int, default=4, help="Number of frames to skip between images"),
    ExtraOption("framerate", type=int, default=60, help="Frames per second"),
    ExtraOption("frame_files", nargs="+", default=[], help="Additional frames to import"),
    ExtraOption(
        "color_mode",
        default="nearest",
        choices=["nearest", "exact"],
        help="How to quantize colors." +
             "nearest will map each color to the most similar in the palette." +
             " exact will only match exact colors"
    ),
])
def import_raster(filenames, n_colors, palette, mode, frame_delay=1, framerate=60, frame_files=[], color_mode="nearest"):
    if not isinstance(filenames, list):
        filenames = [filenames]
    filenames = filenames + frame_files

    if mode == "bezier":
        from ..parsers.raster import QuanzationMode
        # TODO QuanzationMode for raster
        cm = QuanzationMode.Nearest if ns.color_mode == "nearest" else QuanzationMode.Exact

        return raster_to_animation(
            filenames, n_colors, frame_delay,
            framerate=framerate,
            palette=palette,
            mode=cm
        )
    elif mode == "polygon":
        return pixel_to_animation_paths(filenames, frame_delay, framerate)
    else:
        return pixel_to_animation(filenames, frame_delay, framerate)




from .base import TgsEnum


## \ingroup Lottie
class TestBased(TgsEnum):
    Characters = 1
    CharacterExcludingSpaces = 2
    Words = 3
    Lines = 4

    @classmethod
    def default(cls):
        return cls.Characters


## \ingroup Lottie
class TextShape(TgsEnum):
    Square = 1
    RampUp = 2
    RampDown = 3
    Triangle = 4
    Round = 5
    Smooth = 6

    @classmethod
    def default(cls):
        return cls.Square


## \ingroup Lottie
class TextGrouping(TgsEnum):
    Characters = 1
    Word = 2
    Line = 3
    All = 4

    @classmethod
    def default(cls):
        return cls.Characters



import re
import math
from xml.etree import ElementTree

from .handler import SvgHandler, NameMode
from ... import objects
from ...nvector import NVector
from ...utils import restructure
from ...utils.transform import TransformMatrix
try:
    from ...utils import font
    has_font = True
except ImportError:
    has_font = False


class PrecompTime:
    def __init__(self, pcl: objects.PreCompLayer):
        self.pcl = pcl

    def get_time_offset(self, time, lot):
        remap = time
        if self.pcl.time_remapping:
            remapf = self.pcl.time_remapping.get_value(time)
            remap = lot.in_point * (1-remapf) + lot.out_point * remapf

        return remap - self.pcl.start_time


class SvgBuilder(SvgHandler, restructure.AbstractBuilder):
    merge_paths = True
    namestart = (
        r":_A-Za-z\xC0-\xD6\xD8-\xF6\xF8-\u02FF\u0370-\u037D\u037F-\u1FFF" +
        r"\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF" +
        r"\uFDF0-\uFFFD\U00010000-\U000EFFFF"
    )
    namenostart = r"-.0-9\xB7\u0300-\u036F\u203F-\u2040"
    id_re = re.compile("^[%s][%s%s]*$" % (namestart, namenostart, namestart))

    def __init__(self, time=0):
        super().__init__()
        self.svg = ElementTree.Element("svg")
        self.dom = ElementTree.ElementTree(self.svg)
        self.svg.attrib["xmlns"] = self.ns_map["svg"]
        self.ids = set()
        self.idc = 0
        self.name_mode = NameMode.Inkscape
        self.actual_time = time
        self.precomp_times = []
        self._precomps = {}
        self._assets = {}
        self._current_layer = []

    @property
    def time(self):
        time = self.actual_time
        if self.precomp_times:
            for pct in self.precomp_times:
                time = pct.get_time_offset(time, self._current_layer[-1])
        return time

    def gen_id(self, prefix="id"):
        while True:
            self.idc += 1
            id = "%s_%s" % (prefix, self.idc)
            if id not in self.ids:
                break
        self.ids.add(id)
        return id

    def set_clean_id(self, dom, n):
        idn = n.replace(" ", "_")
        if self.id_re.match(idn) and idn not in self.ids:
            self.ids.add(idn)
        else:
            idn = self.gen_id(dom.tag)

        dom.attrib["id"] = idn
        return idn

    def set_id(self, dom, lottieobj, inkscape_qual=None, force=False):
        n = getattr(lottieobj, "name", None)
        if n is None or self.name_mode == NameMode.NoName:
            if force:
                id = self.gen_id(dom.tag)
                dom.attrib["id"] = id
                return id
            return None

        idn = self.set_clean_id(dom, n)
        if inkscape_qual:
            dom.attrib[inkscape_qual] = n
        return idn

    def _on_animation(self, animation: objects.Animation):
        self.svg.attrib["width"] = str(animation.width)
        self.svg.attrib["height"] = str(animation.height)
        self.svg.attrib["viewBox"] = "0 0 %s %s" % (animation.width, animation.height)
        self.svg.attrib["version"] = "1.1"
        self.set_id(self.svg, animation, self.qualified("sodipodi", "docname"))
        self.defs = ElementTree.SubElement(self.svg, "defs")
        if self.name_mode == NameMode.Inkscape:
            self.svg.attrib[self.qualified("inkscape", "export-xdpi")] = "96"
            self.svg.attrib[self.qualified("inkscape", "export-ydpi")] = "96"
            namedview = ElementTree.SubElement(self.svg, self.qualified("sodipodi", "namedview"))
            namedview.attrib[self.qualified("inkscape", "pagecheckerboard")] = "true"
            namedview.attrib["borderlayer"] = "true"
            namedview.attrib["bordercolor"] = "#666666"
            namedview.attrib["pagecolor"] = "#ffffff"
        self.svg.attrib["style"] = "fill: none; stroke: none"

        self._current_layer = [animation]
        return self.svg

    def _mask_to_def(self, mask):
        svgmask = ElementTree.SubElement(self.defs, "mask")
        mask_id = self.gen_id()
        svgmask.attrib["id"] = mask_id
        svgmask.attrib["mask-type"] = "alpha"
        path = ElementTree.SubElement(svgmask, "path")
        path.attrib["d"] = self._bezier_to_d(mask.shape.get_value(self.time))
        path.attrib["fill"] = "#fff"
        path.attrib["fill-opacity"] = str(mask.opacity.get_value(self.time) / 100)
        return mask_id

    def _on_masks(self, masks):
        if len(masks) == 1:
            return self._mask_to_def(masks[0])
        mask_ids = list(map(self._mask_to_def, masks))
        mask_def = ElementTree.SubElement(self.defs, "mask")
        mask_id = self.gen_id()
        mask_def.attrib["id"] = mask_id
        g = mask_def
        for mid in mask_ids:
            g = ElementTree.SubElement(g, "g")
            g.attrib["mask"] = "url(#%s)" % mid
        full = ElementTree.SubElement(g, "rect")
        full.attrib["fill"] = "#fff"
        full.attrib["width"] = self.svg.attrib["width"]
        full.attrib["height"] = self.svg.attrib["height"]
        full.attrib["x"] = "0"
        full.attrib["y"] = "0"
        return mask_id

    def _on_layer(self, layer_builder, dom_parent):
        lot = layer_builder.lottie
        self._current_layer.append(lot)

        if not self.precomp_times and (lot.in_point > self.time or lot.out_point < self.time):
            self._current_layer.pop()
            return None

        g = self.group_from_lottie(lot, dom_parent, True)

        if lot.masks:
            g.attrib["mask"] = "url(#%s)" % self._on_masks(lot.masks)

        if isinstance(lot, objects.PreCompLayer):
            self.precomp_times.append(PrecompTime(lot))

            for layer in self._precomps.get(lot.reference_id, []):
                self.process_layer(layer, g)

            self.precomp_times.pop()
        elif isinstance(lot, objects.NullLayer):
            g.attrib["opacity"] = "1"
        elif isinstance(lot, objects.ImageLayer):
            use = ElementTree.SubElement(g, "use")
            use.attrib[self.qualified("xlink", "href")] = "#" + self._assets[lot.image_id]
        elif isinstance(lot, objects.TextLayer):
            self._on_text_layer(g, lot)
        elif isinstance(lot, objects.SolidColorLayer):
            rect = ElementTree.SubElement(g, "rect")
            rect.attrib["width"] = str(lot.width)
            rect.attrib["height"] = str(lot.height)
            rect.attrib["fill"] = lot.color

        if not lot.name:
            g.attrib[self.qualified("inkscape", "label")] = lot.__class__.__name__
        if layer_builder.shapegroup:
            g.attrib["style"] = self.group_to_style(layer_builder.shapegroup)
            self._split_stroke(layer_builder.shapegroup, g, dom_parent)
        if lot.hidden:
            g.attrib.setdefault("style", "")
            g.attrib["style"] += "display: none;"

        return g

    def _on_text_layer(self, g, lot):
        text = ElementTree.SubElement(g, "text")
        doc = lot.data.get_value(self.time)
        if doc:
            text.attrib["font-family"] = doc.font_family
            text.attrib["font-size"] = str(doc.font_size)
            if doc.line_height:
                text.attrib["line-height"] = "%s%%" % doc.line_height
            if doc.justify == objects.text.TextJustify.Left:
                text.attrib["text-align"] = "start"
            elif doc.justify == objects.text.TextJustify.Center:
                text.attrib["text-align"] = "center"
            elif doc.justify == objects.text.TextJustify.Right:
                text.attrib["text-align"] = "end"

            text.attrib["fill"] = color_to_css(doc.color)
            text.text = doc.text

    def _on_layer_end(self, out_layer):
        self._current_layer.pop()

    def _on_precomp(self, id, dom_parent, layers):
        self._precomps[id] = layers

    def _on_asset(self, asset):
        if isinstance(asset, objects.assets.Image):
            img = ElementTree.SubElement(self.defs, "image")
            xmlid = self.set_clean_id(img, asset.id)
            self._assets[asset.id] = xmlid
            img.attrib[self.qualified("xlink", "href")] = asset.image or asset.image_path
            img.attrib["width"] = str(asset.width)
            img.attrib["height"] = str(asset.height)

    def _get_value(self, prop, default=NVector(0, 0)):
        if prop:
            v = prop.get_value(self.time)
        else:
            v = default

        if v is None:
            return default
        if isinstance(v, NVector):
            return v.clone()
        return v

    def set_transform(self, dom, transform):
        trans = []
        mat = TransformMatrix()

        anchor = self._get_value(transform.anchor_point)
        mat.translate(-anchor.x, -anchor.y)

        scale = self._get_value(transform.scale, NVector(100, 100))
        mat.scale(scale.x / 100, scale.y / 100)

        skew = self._get_value(transform.skew, 0) * math.pi / 180
        if skew != 0:
            axis = self._get_value(transform.skew_axis, 0) * math.pi / 180
            mat.skew_from_axis(-skew, axis)

        rot = self._get_value(transform.rotation, 0) * math.pi / 180
        if rot:
            mat.rotate(-rot)

        # TODO auto orient

        pos = self._get_value(transform.position)
        mat.translate(pos.x, pos.y)

        dom.attrib["transform"] = mat.to_css_2d()

        if transform.opacity is not None:
            op = transform.opacity.get_value(self.time)
            if op != 100:
                dom.attrib["opacity"] = str(op/100)

    def _get_group_stroke(self, group):
        style = {}
        if group.stroke:
            if isinstance(group.stroke, objects.GradientStroke):
                style["stroke"] = "url(#%s)" % self.process_gradient(group.stroke)
            else:
                style["stroke"] = color_to_css(group.stroke.color.get_value(self.time))

            style["stroke-opacity"] = group.stroke.opacity.get_value(self.time) / 100
            style["stroke-width"] = group.stroke.width.get_value(self.time)
            if group.stroke.miter_limit is not None:
                style["stroke-miterlimit"] = group.stroke.miter_limit

            if group.stroke.line_cap == objects.LineCap.Round:
                style["stroke-linecap"] = "round"
            elif group.stroke.line_cap == objects.LineCap.Butt:
                style["stroke-linecap"] = "butt"
            elif group.stroke.line_cap == objects.LineCap.Square:
                style["stroke-linecap"] = "square"

            if group.stroke.line_join == objects.LineJoin.Round:
                style["stroke-linejoin"] = "round"
            elif group.stroke.line_join == objects.LineJoin.Bevel:
                style["stroke-linejoin"] = "bevel"
            elif group.stroke.line_join == objects.LineJoin.Miter:
                style["stroke-linejoin"] = "miter"

            if group.stroke.dashes:
                dasharray = []
                last = 0
                last_mode = objects.StrokeDashType.Dash
                for dash in group.stroke.dashes:
                    if last_mode == dash.type:
                        last += dash.length.get_value(self.time)
                    else:
                        if last_mode != objects.StrokeDashType.Offset:
                            dasharray.append(str(last))
                            last = 0
                        last_mode = dash.type
                style["stroke-dasharray"] = " ".join(dasharray)
        return style

    def _style_to_css(self, style):
        return ";".join(map(
            lambda x: ":".join(map(str, x)),
            style.items()
        ))

    def _split_stroke(self, group, fill_layer, out_parent):
        if not group.stroke:# or group.stroke_above:
            return

        style = self._get_group_stroke(group)
        if style.get("stroke-width", 0) <= 0 or style["stroke-opacity"] <= 0:
            return

        g = ElementTree.Element("g")
        self.set_clean_id(g, "stroke")
        use = ElementTree.Element("use")
        for i, e in enumerate(out_parent):
            if e is fill_layer:
                out_parent.insert(i, g)
                out_parent.remove(fill_layer)
                break
        else:
            return

        if group.stroke_above:
            g.append(fill_layer)
            g.append(use)
        else:
            g.append(use)
            g.append(fill_layer)

        use.attrib[self.qualified("xlink", "href")] = "#" + fill_layer.attrib["id"]
        use.attrib["style"] = self._style_to_css(style)
        return g

    def group_to_style(self, group):
        style = {}
        if group.fill:
            style["fill-opacity"] = group.fill.opacity.get_value(self.time) / 100
            if isinstance(group.fill, objects.GradientFill):
                style["fill"] = "url(#%s)" % self.process_gradient(group.fill)
            else:
                style["fill"] = color_to_css(group.fill.color.get_value(self.time))

            if group.fill.fill_rule:
                style["fill-rule"] = "evenodd" if group.fill.fill_rule == objects.FillRule.EvenOdd else "nonzero"

        #if group.stroke_above:
            #style.update(self._get_group_stroke(group))

        return self._style_to_css(style)

    def process_gradient(self, gradient):
        spos = gradient.start_point.get_value(self.time)
        epos = gradient.end_point.get_value(self.time)

        if gradient.gradient_type == objects.GradientType.Linear:
            dom = ElementTree.SubElement(self.defs, "linearGradient")
            dom.attrib["x1"] = str(spos[0])
            dom.attrib["y1"] = str(spos[1])
            dom.attrib["x2"] = str(epos[0])
            dom.attrib["y2"] = str(epos[1])
        elif gradient.gradient_type == objects.GradientType.Radial:
            dom = ElementTree.SubElement(self.defs, "radialGradient")
            dom.attrib["cx"] = str(spos[0])
            dom.attrib["cy"] = str(spos[1])
            dom.attrib["r"] = str((epos-spos).length)
            a = gradient.highlight_angle.get_value(self.time) * math.pi / 180
            l = gradient.highlight_length.get_value(self.time)
            dom.attrib["fx"] = str(spos[0] + math.cos(a) * l)
            dom.attrib["fy"] = str(spos[1] + math.sin(a) * l)

        id = self.set_id(dom, gradient, force=True)
        dom.attrib["gradientUnits"] = "userSpaceOnUse"

        for off, color in gradient.colors.stops_at(self.time):
            stop = ElementTree.SubElement(dom, "stop")
            stop.attrib["offset"] = "%s%%" % (off * 100)
            stop.attrib["stop-color"] = color_to_css(color[:3])
            if len(color) > 3:
                stop.attrib["stop-opacity"] = str(color[3])

        return id

    def group_from_lottie(self, lottie, dom_parent, layer):
        g = ElementTree.SubElement(dom_parent, "g")
        if layer and self.name_mode == NameMode.Inkscape:
            g.attrib[self.qualified("inkscape", "groupmode")] = "layer"
        self.set_id(g, lottie, self.qualified("inkscape", "label"), force=True)
        self.set_transform(g, lottie.transform)
        return g

    def _on_shapegroup(self, group, dom_parent):
        if group.empty():
            return

        if len(group.children) == 1 and isinstance(group.children[0], restructure.RestructuredPathMerger):
            path = self.build_path(group.paths.paths, dom_parent)
            self.set_id(path, group.paths.paths[0], force=True)
            path.attrib["style"] = self.group_to_style(group)
            self.set_transform(path, group.lottie.transform)
            return self._split_stroke(group, path, dom_parent)

        g = self.group_from_lottie(group.lottie, dom_parent, group.layer)
        g.attrib["style"] = self.group_to_style(group)
        self.shapegroup_process_children(group, g)
        return self._split_stroke(group, g, dom_parent)

    def _on_merged_path(self, shape, shapegroup, out_parent):
        path = self.build_path(shape.paths, out_parent)
        self.set_id(path, shape.paths[0])
        path.attrib["style"] = self.group_to_style(shapegroup)
        #self._split_stroke(shapegroup, path, out_parent)
        return path

    def _on_shape(self, shape, shapegroup, out_parent):
        if isinstance(shape, objects.Rect):
            svgshape = self.build_rect(shape, out_parent)
        elif isinstance(shape, objects.Ellipse):
            svgshape = self.build_ellipse(shape, out_parent)
        elif isinstance(shape, objects.Star):
            svgshape = self.build_path([shape.to_bezier()], out_parent)
        elif isinstance(shape, objects.Path):
            svgshape = self.build_path([shape], out_parent)
        elif has_font and isinstance(shape, font.FontShape):
            svgshape = self.build_text(shape, out_parent)
        else:
            return
        self.set_id(svgshape, shape, force=True)
        if "style" not in svgshape.attrib:
            svgshape.attrib["style"] = ""
        svgshape.attrib["style"] += self.group_to_style(shapegroup)
        #self._split_stroke(shapegroup, svgshape, out_parent)

        if shape.hidden:
            svgshape.attrib["style"] += "display: none;"
        return svgshape

    def build_rect(self, shape, parent):
        rect = ElementTree.SubElement(parent, "rect")
        size = shape.size.get_value(self.time)
        pos = shape.position.get_value(self.time)
        rect.attrib["width"] = str(size[0])
        rect.attrib["height"] = str(size[1])
        rect.attrib["x"] = str(pos[0] - size[0] / 2)
        rect.attrib["y"] = str(pos[1] - size[1] / 2)
        rect.attrib["rx"] = str(shape.rounded.get_value(self.time))
        return rect

    def build_ellipse(self, shape, parent):
        ellipse = ElementTree.SubElement(parent, "ellipse")
        size = shape.size.get_value(self.time)
        pos = shape.position.get_value(self.time)
        ellipse.attrib["rx"] = str(size[0] / 2)
        ellipse.attrib["ry"] = str(size[1] / 2)
        ellipse.attrib["cx"] = str(pos[0])
        ellipse.attrib["cy"] = str(pos[1])
        return ellipse

    def build_path(self, shapes, parent):
        path = ElementTree.SubElement(parent, "path")
        d = ""
        for shape in shapes:
            bez = shape.shape.get_value(self.time)
            if isinstance(bez, list):
                bez = bez[0]
            if not bez.vertices:
                continue
            if d:
                d += "\n"
            d += self._bezier_to_d(bez)

        path.attrib["d"] = d
        return path

    def _bezier_to_d(self, bez):
        d = "M %s,%s " % tuple(bez.vertices[0].components[:2])
        for i in range(1, len(bez.vertices)):
            qfrom = bez.vertices[i-1]
            h1 = bez.out_tangents[i-1] + qfrom
            qto = bez.vertices[i]
            h2 = bez.in_tangents[i] + qto
            d += "C %s,%s %s,%s %s,%s " % (
                h1[0], h1[1],
                h2[0], h2[1],
                qto[0], qto[1],
            )
        if bez.closed:
            qfrom = bez.vertices[-1]
            h1 = bez.out_tangents[-1] + qfrom
            qto = bez.vertices[0]
            h2 = bez.in_tangents[0] + qto
            d += "C %s,%s %s,%s %s,%s Z" % (
                h1[0], h1[1],
                h2[0], h2[1],
                qto[0], qto[1],
            )

        return d

    def _on_shape_modifier(self, shape, shapegroup, out_parent):
        if isinstance(shape.lottie, objects.Repeater):
            svgshape = self.build_repeater(shape.lottie, shape.child, shapegroup, out_parent)
        elif isinstance(shape.lottie, objects.RoundedCorners):
            svgshape = self.build_rouded_corners(shape.lottie, shape.child, shapegroup, out_parent)
        elif isinstance(shape.lottie, objects.Trim):
            svgshape = self.build_trim_path(shape.lottie, shape.child, shapegroup, out_parent)
        else:
            return self.shapegroup_process_child(shape.child, shapegroup, out_parent)
        return svgshape

    def build_repeater(self, shape, child, shapegroup, out_parent):
        original = self.shapegroup_process_child(child, shapegroup, out_parent)
        if not original:
            return

        ncopies = int(round(shape.copies.get_value(self.time)))
        if ncopies == 1:
            return

        out_parent.remove(original)

        g = ElementTree.SubElement(out_parent, "g")
        self.set_clean_id(g, "repeater")

        for copy in range(ncopies-1):
            use = ElementTree.SubElement(g, "use")
            use.attrib[self.qualified("xlink", "href")] = "#" + original.attrib["id"]

        orig_wrapper = ElementTree.SubElement(g, "g")
        orig_wrapper.append(original)

        transform = objects.Transform()
        so = shape.transform.start_opacity.get_value(self.time)
        eo = shape.transform.end_opacity.get_value(self.time)
        position = shape.transform.position.get_value(self.time)
        rotation = shape.transform.rotation.get_value(self.time)
        anchor_point = shape.transform.anchor_point.get_value(self.time)
        for i in range(ncopies-1, -1, -1):
            of = i / (ncopies-1)
            transform.opacity.value = so * of + eo * (1 - of)
            self.set_transform(g[i], transform)
            transform.position.value += position
            transform.rotation.value += rotation
            transform.anchor_point.value += anchor_point

        return g

    def build_rouded_corners(self, shape, child, shapegroup, out_parent):
        round_amount = shape.radius.get_value(self.time)
        return self._modifier_process(child, shapegroup, out_parent, self._build_rouded_corners_shape, round_amount)

    def _build_rouded_corners_shape(self, shape, round_amount):
        if not isinstance(shape, objects.Shape):
            return [shape]
        path = sh.to_bezier()
        bezier = path.shape.get_value(self.time).rounded(round_amount)
        path.shape.clear_animation(bezier)
        return [path]

    def build_trim_path(self, shape, child, shapegroup, out_parent):
        start = max(0, min(1, shape.start.get_value(self.time) / 100))
        end = max(0, min(1, shape.end.get_value(self.time) / 100))
        offset = shape.offset.get_value(self.time) / 360 % 1

        multidata = {}
        length = 0

        if shape.multiple == objects.TrimMultipleShapes.Individually:
            for visishape in reversed(list(self._modifier_foreach_shape(child))):
                bez = visishape.to_bezier().shape.get_value(self.time)
                local_length = bez.rough_length()
                multidata[visishape] = (bez, length, local_length)
                length += local_length

        return self._modifier_process(
            child, shapegroup, out_parent, self._build_trim_path_shape,
            start+offset, end+offset, multidata, length
        )

    def _modifier_foreach_shape(self, shape):
        if isinstance(shape, restructure.RestructuredShapeGroup):
            for child in shape.children:
                for chsh in self._modifier_foreach_shape(child):
                    yield chsh
        elif isinstance(shape, restructure.RestructuredPathMerger):
            for p in shape.paths:
                yield p
        elif isinstance(shape, objects.Shape):
            yield shape

    def _modifier_process(self, child, shapegroup, out_parent, callback, *args):
        children = self._modifier_process_child(child, shapegroup, out_parent, callback, *args)
        return [self.shapegroup_process_child(ch, shapegroup, out_parent) for ch in children]

    def _trim_offlocal(self, t, local_start, local_length, total_length):
        gt = (t * total_length - local_start) / local_length
        return max(0, min(1, gt))

    def _build_trim_path_shape(self, shape, start, end, multidata, total_length):
        if not isinstance(shape, objects.Shape):
            return [shape]

        if multidata:
            bezier, local_start, local_length = multidata[shape]
            if end > 1:
                lstart = self._trim_offlocal(start, local_start, local_length, total_length)
                lend = self._trim_offlocal(end-1, local_start, local_length, total_length)
                out = []
                if lstart < 1:
                    out.append(objects.Path(bezier.segment(lstart, 1)))
                if lend > 0:
                    out.append(objects.Path(bezier.segment(0, lend)))
                return out

            lstart = self._trim_offlocal(start, local_start, local_length, total_length)
            lend = self._trim_offlocal(end, local_start, local_length, total_length)
            if lend <= 0 or lstart >= 1:
                return []
            if lstart <= 0 and lend >= 1:
                return [objects.Path(bezier)]
            seg = bezier.segment(lstart, lend)
            return [objects.Path(seg)]

        path = shape.to_bezier()
        bezier = path.shape.get_value(self.time)
        if end > 1:
            bez1 = bezier.segment(start, 1)
            bez2 = bezier.segment(0, end-1)
            return [objects.Path(bez1), objects.Path(bez2)]
        else:
            seg = bezier.segment(start, end)
            return [objects.Path(seg)]

    def _modifier_process_children(self, shapegroup, out_parent, callback, *args):
        children = []
        for shape in shapegroup.children:
            children.extend(self._modifier_process_child(shape, shapegroup, out_parent, callback, *args))
        shapegroup.children = children

    def _modifier_process_child(self, shape, shapegroup, out_parent, callback, *args):
        if isinstance(shape, restructure.RestructuredShapeGroup):
            self._modifier_process_children(shape, out_parent, callback, *args)
            return [shape]
        elif isinstance(shape, restructure.RestructuredPathMerger):
            paths = []
            for p in shape.paths:
                paths.extend(callback(p, *args))
            shape.paths = paths
            if paths:
                return [shape]
            return []
        else:
            return callback(shape, *args)

    def _custom_object_supported(self, shape):
        if has_font and isinstance(shape, font.FontShape):
            return True
        return False

    def build_text(self, shape, parent):
        text = ElementTree.SubElement(parent, "text")
        if "family" in shape.query:
            text.attrib["font-family"] = shape.query["family"]
        if "weight" in shape.query:
            text.attrib["font-weight"] = str(shape.query.weight_to_css())
        slant = int(shape.query.get("slant", 0))
        if slant > 0 and slant < 110:
            text.attrib["font-style"] = "italic"
        elif slant >= 110:
            text.attrib["font-style"] = "oblique"

        text.attrib["font-size"] = str(shape.size)

        text.attrib["white-space"] = "pre"

        pos = shape.style.position
        text.attrib["x"] = str(pos.x)
        text.attrib["y"] = str(pos.y)
        text.text = shape.text

        return text


def color_to_css(color):
    #if len(color) == 4:
        #return ("rgba(%s, %s, %s" % tuple(map(lambda c: int(round(c*255)), color[:3]))) + ", %s)" % color[3]
    return "rgb(%s, %s, %s)" % tuple(map(lambda c: int(round(c*255)), color[:3]))


def to_svg(animation, time):
    builder = SvgBuilder(time)
    builder.process(animation)
    return builder.dom



from . import animation, ellipse, ik, linediff, restructure, script, stripper
__all__ = ["animation", "ellipse", "ik", "linediff", "restructure", "script", "stripper"]

try:
    from . import font
    __all__ += ["font"]
except ImportError:
    pass



from xml.dom import minidom
import enum

from tgs.nvector import NVector
from tgs.parsers.sif.ast_impl.base import SifAstNode, TypeDescriptor, ObjectRegistry
from tgs.parsers.sif.xml.animatable import XmlAnimatable
from tgs.parsers.sif.xml.wrappers import XmlBoneReference, XmlSifElement, XmlList
from tgs.parsers.sif.sif.nodes import Segment, WeightedVector, Bline
from tgs.parsers.sif.sif.enums import Smooth
from tgs.parsers.sif.sif.core import SifNodeMeta, FrameTime


class SifAstComplex(SifAstNode, metaclass=SifNodeMeta):
    _nodes = []

    def __init__(self, **kw):
        for node in self._nodes:
            node.initialize_object(kw, self)

    @classmethod
    def from_dom(cls, xml: minidom.Element, param: TypeDescriptor, registry: ObjectRegistry):
        instance = cls.get_class_from_dom(xml, param, registry)()
        for node in cls._nodes:
            node.from_xml(instance, xml, registry, param)
        return instance

    @classmethod
    def get_class_from_dom(cls, xml: minidom.Element, param: TypeDescriptor, registry: ObjectRegistry):
        return cls

    def to_dom(self, dom: minidom.Document, param: TypeDescriptor):
        element = self._prepare_to_dom(dom, param)
        for node in self._nodes:
            node.to_xml(self, element, dom, param)
        return element


class SifAstBoneLink(SifAstComplex):
    _tag = "bone_link"

    _nodes = [
        XmlBoneReference("bone"),
        XmlAnimatable("base_value", "vector", NVector(0, 0)),
        XmlAnimatable("translate", "bool", True),
        XmlAnimatable("rotate", "bool", True),
        XmlAnimatable("skew", "bool", True),
        XmlAnimatable("scale_x", "bool", True),
        XmlAnimatable("scale_y", "bool", True),
    ]


class SifAstBoneInfluence(SifAstComplex):
    _tag = "boneinfluence"

    _nodes = [
        # TODO bone_weight_list
        XmlAnimatable("link", "vector", NVector(0, 0)),
    ]


class SifSegCalcTangent(SifAstComplex):
    _tag = "segcalctangent"

    _nodes = [
        XmlSifElement("segment", Segment),
        XmlAnimatable("amount", "real", .5),
    ]


class SifSegCalcVertex(SifAstComplex):
    _tag = "segcalcvertex"

    _nodes = [
        XmlSifElement("segment", Segment),
        XmlAnimatable("amount", "real", .5),
    ]


class WeightedAverage(SifAstComplex):
    _tag = "weighted_average"

    _nodes = [
        XmlList(WeightedVector, "vectors", "entry"),
    ]

    def _prepare_to_dom(self, dom: minidom.Document, param: TypeDescriptor):
        element = dom.createElement(self._tag)
        element.setAttribute("type", "weighted_vector")
        return element


class SifAdd(SifAstComplex):
    _tag = "add"

    _nodes = [
        XmlAnimatable("lhs", "_recurse"),
        XmlAnimatable("rhs", "_recurse"),
        XmlAnimatable("scalar", "real", 1.),
    ]


class SifAnimatedFile(SifAstComplex):
    _tag = "animated_file"

    _nodes = [
        XmlAnimatable("filename", "string"),
    ]


class SifAnimatedFile(SifAstComplex):
    _tag = "animated_file"

    _nodes = [
        XmlAnimatable("filename", "string"),
    ]


class Accuracy(enum.Enum):
    Rough = 0
    Normal = 1
    Fine = 2
    Extreme = 3


class DerivativeOrder:
    FirstDerivative = 0
    SecondDerivative = 1


class SifDerivative(SifAstComplex):
    _tag = "derivative"

    _nodes = [
        XmlAnimatable("link", "_recurse"),
        XmlAnimatable("interval", "real", 0.01),
        XmlAnimatable("accuracy", "integer", Accuracy.Normal, Accuracy),
        XmlAnimatable("order", "integer", DerivativeOrder.FirstDerivative, DerivativeOrder),
    ]


class SifDynamic(SifAstComplex):
    _tag = "dynamic"

    _nodes = [
        XmlAnimatable("tip_static", "vector", NVector(0, 0)),
        XmlAnimatable("origin", "vector", NVector(0, 0)),
        XmlAnimatable("force", "vector", NVector(0, 0)),
        XmlAnimatable("torque", "real", 0.),
        XmlAnimatable("damping", "real", 0.4),
        XmlAnimatable("friction", "real", 0.4),
        XmlAnimatable("spring", "real", 30.),
        XmlAnimatable("torsion", "real", 30.),
        XmlAnimatable("mass", "real", 0.3),
        XmlAnimatable("inertia", "real", 0.3),
        XmlAnimatable("spring_rigid", "bool", False),
        XmlAnimatable("torsion_rigid", "bool", False),
        XmlAnimatable("origin_drags_tip", "bool", True),
    ]


class SifGreyed(SifAstComplex):
    _tag = "greyed"

    _nodes = [
        XmlAnimatable("link", "_recurse"),
    ]


class SifLinear(SifAstComplex):
    _tag = "linear"

    _nodes = [
        XmlAnimatable("slope", "vector", NVector(0, 0)),
        XmlAnimatable("offset", "vector", NVector(0, 0)),
    ]


class SifRadialComposite(SifAstComplex):
    _tag = "radial_composite"

    _nodes = [
        XmlAnimatable("radius", "real", 0.),
        XmlAnimatable("theta", "angle", 0.),
    ]


class SifComposite(SifAstComplex):
    _tag = "composite"

    @classmethod
    def get_class_from_dom(cls, xml: minidom.Element, param: TypeDescriptor, registry: ObjectRegistry):
        type = xml.getAttribute("type")
        if type == "vector":
            return SifVectorComposite
        return None

    def _prepare_to_dom(self, dom: minidom.Document, param: TypeDescriptor):
        element = dom.createElement("composite")
        element.setAttribute("type", param.typename)
        return element


class SifVectorComposite(SifComposite):
    _type = "vector"

    _nodes = [
        XmlAnimatable("x", "real", 0.),
        XmlAnimatable("y", "real", 0.),
    ]


class SifRandom(SifAstComplex):
    _tag = "random"

    _nodes = [
        XmlAnimatable("link", "_recurse"),
        XmlAnimatable("radius", "real", 0.),
        XmlAnimatable("seed", "integer", 0),
        XmlAnimatable("speed", "real", 1.),
        XmlAnimatable("smooth", "integer", Smooth.Cubic, Smooth),
        XmlAnimatable("loop", "real", 0.),
    ]


class SifReference(SifAstComplex):
    _tag = "link"

    _nodes = [
        XmlAnimatable("reference", "_recurse"),
    ]


class SifScale(SifAstComplex):
    _tag = "scale"

    _nodes = [
        XmlAnimatable("link", "_recurse"),
        XmlAnimatable("scalar", "real", 1.),
    ]


class SifStep(SifAstComplex):
    _tag = "step"

    _nodes = [
        XmlAnimatable("link", "_recurse"),
        XmlAnimatable("duration", "time", FrameTime(1, FrameTime.Unit.Seconds)),
        XmlAnimatable("start_time", "time", FrameTime(0, FrameTime.Unit.Seconds)),
        XmlAnimatable("intersection", "real", 0.5),
    ]


class SifSubtract(SifAstComplex):
    _tag = "subtract"

    _nodes = [
        XmlAnimatable("lhs", "_recurse"),
        XmlAnimatable("rhs", "_recurse"),
        XmlAnimatable("scalar", "real", 1.),
    ]


class SifSwitch(SifAstComplex):
    _tag = "switch"

    _nodes = [
        XmlAnimatable("link_off", "_recurse"),
        XmlAnimatable("link_on", "_recurse"),
        XmlAnimatable("switch", "bool", False),
    ]


class SifTimedSwap(SifAstComplex):
    _tag = "timed_swap"

    _nodes = [
        XmlAnimatable("before", "_recurse"),
        XmlAnimatable("after", "_recurse"),
        XmlAnimatable("time", "time", FrameTime(0, FrameTime.Unit.Seconds)),
        XmlAnimatable("length", "time", FrameTime(0, FrameTime.Unit.Seconds)),
    ]


class SifTimeLoop(SifAstComplex):
    _tag = "timeloop"

    _nodes = [
        XmlAnimatable("link", "_recurse"),
        XmlAnimatable("link_time", "time", FrameTime(0, FrameTime.Unit.Seconds)),
        XmlAnimatable("local_time", "time", FrameTime(0, FrameTime.Unit.Seconds)),
        XmlAnimatable("duration", "time", FrameTime(0, FrameTime.Unit.Seconds)),
    ]


class SifPower(SifAstComplex):
    _tag = "power"

    _nodes = [
        XmlAnimatable("base", "real", 1.),
        XmlAnimatable("power", "real", 1.),
        XmlAnimatable("epsilon", "real", 0.000001),
        XmlAnimatable("infinite", "real", 999999.),
    ]



class SifBlineCalcTangent(SifAstComplex):
    _tag = "blinecalctangent"

    _nodes = [
        XmlSifElement("bline", Bline),
        XmlAnimatable("loop", "bool", False),
        XmlAnimatable("amount", "real", 0.5),
        XmlAnimatable("offset", "angle", 0.),
        XmlAnimatable("scale", "real", 1.),
        XmlAnimatable("fixed_length", "bool", False),
        XmlAnimatable("homogeneous", "bool", False),
    ]


class SifBlineCalcVertex(SifAstComplex):
    _tag = "blinecalcvertex"

    _nodes = [
        XmlSifElement("bline", Bline),
        XmlAnimatable("loop", "bool", False),
        XmlAnimatable("amount", "real", 0.5),
        XmlAnimatable("homogeneous", "bool", False),
    ]
